"""
 A cache of MD5 meta data associated with VOSpace transfers.
 
 When transferring large numbers of files to and from VOSpace there is a substantial expectation of network failures 
 which will require that a transfer be re-attempted.  This module provides a class that keeps track of the MD5 values
 associated files on disk and in VOSpace allowing the caller to choose to skip files that match (MD5 wise) between the 
 two locations.  
"""
import sqlite3
import logging
import hashlib
import tempfile

READ_BUFFER_SIZE = 8192


class MD5Cache:

    def __init__(self, cache_db=None):
        """Setup the sqlDB that will contain the cache table.
        
        The slqDB can then be used to lookup MD5 values rather than recompute them at each restart of a transfer.
        
        :param cache_db: The path and filename where the SQL db will be stored. 
        """
        if cache_db is None:
            self.cache_obj = tempfile.NamedTemporaryFile()
            self.cache_db = self.cache_obj.name
        else:
            self.cache_db = cache_db

        # initialize the md5Cache db
        sql_conn = sqlite3.connect(self.cache_db)
        with sql_conn:
            # build cache lookup if doesn't already exists
            sql_conn.execute(("create table if not exists "
                             "md5_cache (filename text PRIMARY KEY NOT NULL , md5 text, st_size int, st_mtime int)"))

    @staticmethod
    def compute_md5(filename, block_size=READ_BUFFER_SIZE):
        """
        A convenience routine that computes and returns the MD5 of the supplied filename.
        :param filename: Name of the file to compute the MD5 checksum for.
        :param block_size: Loop through the file with this number of bytes per call.
        :return: the MD5 hexdigest of the file.
        :rtype: str
        """
        md5 = hashlib.md5()
        with open(filename, 'rb') as f:
            while True:
                buf = f.read(block_size)
                if len(buf) == 0:
                    break
                md5.update(buf)
        return md5.hexdigest()

    def get(self, filename):
        """Get the MD5 for filename.  
        
        First look in MD5 cache databse and then compute if needbe.
        
        :param filename: name of the file you want the MD5 sum for.
        """
        slq_conn = sqlite3.connect(self.cache_db)
        with slq_conn:
            cursor = slq_conn.execute("SELECT md5, st_size, st_mtime FROM md5_cache WHERE filename = ? ", (filename,))
            md5_row = cursor.fetchone()
        if md5_row is not None:
            return md5_row
        else:
            return None

    def delete(self, filename):
        """Delete a record from the cache MD5 database.
        
        :param filename: Name of the file whose md5 record to be deleted from the cache database
        """
        sql_conn = sqlite3.connect(self.cache_db)
        with sql_conn:
            sql_conn.execute("DELETE from md5_cache WHERE filename = ?", (filename,))

    def update(self, filename, md5, st_size, st_mtime):
        """Update the MD5 value stored in the cache db
        
        :param filename: Name of the file to update the MD5 value for.
        :param md5: the MD5 hexdigest to be stored to the databse.
        :param st_size: size of the file being updated (stored to database)
        :param st_mtime: last modified time of the file being stored to database.
        """

        # UPDATE the MD5 database
        sql_connection = sqlite3.connect(self.cache_db)
        try:
            with sql_connection:
                sql_connection.execute("DELETE from md5_cache WHERE filename = ?", (filename,))
                sql_connection.execute(("INSERT INTO md5_cache (filename, md5, st_size, st_mtime)"
                                 " VALUES ( ?, ?, ?, ?)"), (filename, md5, st_size, st_mtime))
        except Exception as e:
            logging.error(e)
        return md5



#!python
"""Create a directory (ContainerNode) in the VOSpace repository"""

import os
import logging
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception
from .. import vos

DESCRIPTION = """creates a new VOSpace ContainerNode (aka directory).

eg vmkdir vos:RootNode/NewContiner"""


def vmkdir():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument('container_node', action='store', help='Name of the container node to craete.')
    parser.add_argument("-p", action="store_true", help="Create intermediate directories as required.")

    args = parser.parse_args()

    set_logging_level_from_args(args)

    logging.info("Creating ContainerNode (directory) {}".format(args.container_node))

    try:

        client = vos.Client(vospace_certfile=args.certfile, vospace_token=args.token)

        dir_names = []
        this_dir = args.container_node
        if args.p:
            while not client.access(this_dir):
                dir_names.append(os.path.basename(this_dir))
                this_dir = os.path.dirname(this_dir)
            while len(dir_names) > 0:
                this_dir = os.path.join(this_dir, dir_names.pop())
                client.mkdir(this_dir)
        else:
            client.mkdir(this_dir)

    except Exception as ex:
        exit_on_exception(ex)

vmkdir.__doc__ = DESCRIPTION



"""link one VOSpace Node to another."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception
from .. import vos
from argparse import ArgumentError

DESCRIPTION = """
vln creates a new VOSpace entry (LinkNode, target) that has the same modes as the source Node.
It is useful for maintaining multiple copies of a Node in many places at once without using
up storage for the ''copies''; instead, a link ''points'' to the original copy.

Only symbolic links are supported.

vln vos:VOSpaceSource vos:VOSpaceTarget


examples:

    vln vos:vospace/junk.txt vos:vospace/linkToJunk.txt
    vln vos:vospace/directory vos:vospace/linkToDirectory
    vln http://external.data.source vos:vospace/linkToExternalDataSource

"""


def vln():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument('source', help="location that link will point to.")
    parser.add_argument('target', help="location of the LinkNode")

    try:
        opt = parser.parse_args()
        set_logging_level_from_args(opt)

        if not (opt.source.startswith('vos:') or opt.source.startswith('http:')) or not opt.target.startswith('vos:'):
            raise ArgumentError(None, "source must be vos node or http url, target must be vos node")

        client = vos.Client(vospace_certfile=opt.certfile, vospace_token=opt.token)
        client.link(opt.source, opt.target)
    except ArgumentError as ex:
        parser.print_usage()
        exit_on_exception(ex)
    except Exception as ex:
        exit_on_exception(ex)

vln.__doc__ = DESCRIPTION


"""A utilities for dealing with CL signal handeling."""
import signal

def signal_handler(signum, frame):
    """
    signal handler for keyboard interupt of cl interface.

    :param signum: signal sent to CL tool. 
    :param frame: frame where CL tool was running
    :raises KeyboardInterrupt
    """
    raise KeyboardInterrupt("SIGNAL {0} from {1} signal handler".format(signum, frame))

signal.signal(signal.SIGINT, signal_handler)




"""cat VOSpace DataNode to stdout"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import sys
import logging
from ..vos import Client
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception


def _cat(uri, cert_filename=None):
    """Cat out the given uri stored in VOSpace.
    
    :param uri: the VOSpace URI that will be piped to stdout.
    :type uri: basestring
    :param cert_filename: filename of the PEM certificate used to gain access.
    :type cert_filename: basestring
    """

    fh = None
    try:
        if uri[0:4] == "vos:":
            fh = Client(vospace_certfile=cert_filename).open(uri, view='data')
            sys.stdout.write(fh.read(return_response=True).text)
        else:
            fh = open(uri, str("r"))
            sys.stdout.write(fh.read())
    finally:
        if fh:
            fh.close()

DESCRIPTION = """Write the content of source (eg. vos:Node/filename) to stdout.

Accepts cutout syntax for FITS files; see vcp --help for syntax details"""


def vcat():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument("source", help="source to cat to stdout out.", nargs="+")
    parser.add_argument("-q", help="run quietly, exit on error without message", action="store_true")

    args = parser.parse_args()
    set_logging_level_from_args(args)

    logger = logging.getLogger()

    exit_code = 0

    try:
        for uri in args.source:
            try:
                _cat(uri, cert_filename=args.certfile)
            except Exception as e:
                exit_code = getattr(e, 'errno', -1)
                if not args.q:
                    logger.error(str(e))
    except KeyboardInterrupt as ke:
        exit_on_exception(ke)

    sys.exit(exit_code)

vcat.__doc__ = DESCRIPTION


#!python
"""A script for sending files to VOSpace via multiple connection streams."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import hashlib
import os
import sys
from multiprocessing import Process, JoinableQueue
from vos.commonparser import CommonParser
import errno
import logging
import time
import signal
from vos import vos, version
from .. import md5_cache


def vsync():
    def signal_handler(signal, frame):
        logger.critical("Interupt\n")
        sys.exit(-1)
    usage = """
      vsync [options] files vos:Destination/

          Version: %s """ % (version.version)
    # handle interupts nicely
    signal.signal(signal.SIGINT, signal_handler)

    startTime = time.time()

    parser = CommonParser(usage)
    parser.add_option('--ignore-checksum', action="store_true", help='dont check MD5 sum, use size and time instead')
    parser.add_option('--cache_nodes', action='store_true', help='cache node MD5 sum in an sqllite db')
    parser.add_option('--recursive', '-r', help="Do a recursive sync", action="store_true")
    parser.add_option('--nstreams', '-n', type=int, help="Number of streams to run (MAX: 30)", default=1)
    parser.add_option('--exclude', help="ignore directories or files containing this pattern", default=None)
    parser.add_option('--include', help="only include files matching this pattern", default=None)
    parser.add_option('--overwrite', help="overwrite copy on server regardless of modification/size/md5 checks", action="store_true")
    parser.add_option('--load_test', action="store_true", help="Used to stress test the VOServer, also set --nstreams to a large value")

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()

    (opt, args) = parser.parse_args()
    parser.process_informational_options()

    logger = logging.getLogger()
    logger.setLevel(parser.log_level)
    logger.addHandler(logging.StreamHandler())

    if len(args) < 2:
        parser.error("requires one or more source files and a single destination directory")

    if opt.nstreams > 30 and not opt.load_test:
        parser.error("Maximum of 30 streams exceeded")

    if opt.cache_nodes:
        from vos import md5_cache
        md5Cache = md5_cache.MD5_Cache()


    dest = args.pop()
    if dest[0:4] != "vos:":
        parser.error("Only allows sync FROM local copy TO VOSpace")
    ## Currently we don't create nodes in sync and we don't sync onto files
    logger.info("Connecting to VOSpace")
    client = vos.Client(vospace_certfile=opt.certfile, vospace_token=opt.token)
    logger.info("Confirming Destination is a directory")
    destIsDir = client.isdir(dest)

    queue = JoinableQueue(maxsize=10 * opt.nstreams)
    goodDirs = []
    nodeDict = {}


    def compute_md5(filename, block_size=None):
        """
        Read through a file and compute that files MD5 checksum.
        :param filename: name of the file on disk
        :param block_size: number of bytes to read into memory, defaults to 2**19 bytes
        :return: md5 as a hexadecimal string
        """
        block_size = block_size is None and 2**19 or block_size
        return md5_cache.MD5Cache.compute_md5(filename, block_size=block_size)

    def fileMD5(filename):
        import os
        md5 = None
        if opt.cache_nodes:
            md5 = md5Cache.get(filename)
        if md5 is None or md5[2] < os.stat(filename).st_mtime:
            md5 = compute_md5(filename)
            if opt.cache_nodes:
                stat = os.stat(filename)
                md5Cache.update(filename, md5, stat.st_size, stat.st_mtime)
        else:
            md5 = md5[0]
        return md5

    class ThreadCopy(Process):
        def __init__(self, queue, client):
            super(ThreadCopy, self).__init__()
            self.client = client
            self.queue = queue
            self.filesSent = 0
            self.filesSkipped = 0
            self.bytesSent = 0
            self.bytesSkipped = 0
            self.filesErrored = 0

        def run(self):
            while True:
                (src, dest) = self.queue.get()
                requeue = (src, dest)
                srcMD5 = None
                stat = os.stat(src)
                if not opt.ignore_checksum and not opt.overwrite:
                    srcMD5 = fileMD5(src)
                if not opt.overwrite:
                    # Check if the file is the same
                    try:
                        nodeInfo = None
                        if opt.cache_nodes:
                            nodeInfo = md5Cache.get(dest)
                        if nodeInfo is None:
                            logger.debug("Getting node info from VOSpace")
                            logger.debug(str(nodeDict.keys()))
                            logger.debug(str(dest))
                            node = self.client.get_node(dest, limit=None)
                            destMD5 = node.props.get('MD5', 'd41d8cd98f00b204e9800998ecf8427e')
                            destLength = node.attr['st_size']
                            destTime = node.attr['st_ctime']
                            if opt.cache_nodes:
                                md5Cache.update(dest, destMD5, destLength, destTime)
                        else:
                            destMD5 = nodeInfo[0]
                            destLength = nodeInfo[1]
                            destTime = nodeInfo[2]
                        logger.debug("Dest MD5: %s " % (destMD5))
                        if (not opt.ignore_checksum and srcMD5 == destMD5) or (opt.ignore_checksum and destTime >= stat.st_mtime and destLength == stat.st_size) :
                            logger.info("skipping: %s  matches %s" % (src, dest))
                            self.filesSkipped += 1
                            self.bytesSkipped += destLength
                            self.queue.task_done()
                            continue
                    except (IOError, OSError) as node_error:
                        """Ignore the erorr"""
                        logger.debug(str(node_error))
                        pass
                logger.info("%s -> %s" % (src, dest))
                try:
                    self.client.copy(src, dest, send_md5=True)
                    node = self.client.get_node(dest, limit=None)
                    destMD5 = node.props.get('MD5', 'd41d8cd98f00b204e9800998ecf8427e')
                    destLength = node.attr['st_size']
                    destTime = node.attr['st_ctime']
                    if opt.cache_nodes:
                           md5Cache.update(dest, destMD5, destLength, destTime)
                    self.filesSent += 1
                    self.bytesSent += stat.st_size
                except (IOError, OSError) as e:
                    logger.error("Error writing %s to server, skipping" % (src))
                    logger.error(str(e))
                    import re
                    if re.search('NodeLocked',str(e)) != None:
                        logger.error("Use vlock to unlock the node before syncing to %s." % (dest))
                    try:
                        if e.errno == 104:
                            self.queue.put(requeue)
                    except Exception as e2:
                        logger.error("Error during requeue")
                        logger.error(str(e2))
                        pass
                    self.filesErrored += 1
                    pass
                self.queue.task_done()


    def mkdirs(dirs):


        logger.debug("%s %s" % (dirs, str(goodDirs)))
        ## if we've seen this before skip it.
        if dirs in goodDirs:
            return

        ## try and make a new directory and return
        ## failure indicates we should see if subdirs exist
        try:
            client.mkdir(dirs)
            logger.info("Made directory %s " % (dirs))
            goodDirs.append(dirs)
            return
        except OSError as e:
            exit_code = getattr(e, 'errno', -1)
            if exit_code != errno.EEXIST:
                raise e

        ## OK, must already have existed, add to list
        goodDirs.append(dirs)

        return


    def copy(source, dest):
        ## strip down dest until we find a part that exists
        ## and then build up the path.  Dest should include the filename
        if os.path.islink(source):
            logger.error("%s is a link, skipping" % (source))
            return
        if not os.access(source, os.R_OK):
            logger.error("Failed to open file %s, skipping" % (source))
            return
        import re
        if re.match('^[A-Za-z0-9\\._\\-\\(\\);:&\\*\\$@!+=\\/]*$', source) is None:
            logger.error("filename %s contains illegal characters, skipping" % (source))
            return

        dirname = os.path.dirname(dest)
        mkdirs(dirname)
        if opt.include is not None and not re.search(opt.include, source):
            return
        queue.put((source, dest), timeout=3600)

    def startStreams(nstreams, vospace_client):
        streams = []
        for i in range(nstreams):
            logger.info("Launching vospace connection stream %d" % (i))
            t = ThreadCopy(queue, client=vospace_client)
            t.daemon = True
            t.start()
            streams.append(t)
        return streams


    def buildFileList(basePath, destRoot='', recursive=False, ignore=None):
        """Build a list of files that should be copied into VOSpace"""
        import string

        spinner = ['-', '\\', '|', '/', '-', '\\', '|', '/']
        count = 0
        import re

        for (root, dirs, filenames) in os.walk(basePath):
            for thisDirname in dirs:
                if not recursive:
                    continue
                thisDirname = os.path.join(root, thisDirname)
                skip = False
                if ignore is not None:
                    for thisIgnore in ignore.split(','):
                        if not thisDirname.find(thisIgnore) < 0:
                            logger.info("excluding: %s " % (thisDirname))
                            skip = True
                            continue
                if skip:
                    continue
                cprefix = os.path.commonprefix((basePath, thisDirname))
                thisDirname = os.path.normpath(destRoot + "/" + thisDirname[len(cprefix):])
                mkdirs(thisDirname)
            for thisfilename in filenames:
                srcfilename = os.path.normpath(os.path.join(root, thisfilename))
                skip = False
                if ignore is not None:
                    for thisIgnore in ignore.split(','):
                        if not srcfilename.find(thisIgnore) < 0:
                            logger.info("excluding: %s " % (srcfilename))
                            skip = True
                            continue
                if skip:
                    continue
                cprefix = os.path.commonprefix((basePath, srcfilename))
                destfilename = os.path.normpath(destRoot + "/" + srcfilename[len(cprefix):])
                thisDirname = os.path.dirname(destfilename)
                mkdirs(thisDirname)

                count += 1
                if opt.verbose:
                    sys.stderr.write("Building list of files to transfer %s\r" % (spinner[count % len(spinner)]))
                copy(srcfilename, destfilename)
            if not recursive:
                return
        return


    streams = startStreams(opt.nstreams, vospace_client=client)

    ### build a complete file list given all the things on the command line
    for filename in args:
        filename = os.path.abspath(filename)
        thisRoot = dest
        if os.path.isdir(filename):
            if filename[-1] != "/" :
                if os.path.basename(filename) != os.path.basename(dest):
                    thisRoot = os.path.join(dest, os.path.basename(filename))
            mkdirs(thisRoot)
            nodeDict[thisRoot] = client.get_node(thisRoot, limit=None)
            try:
                buildFileList(filename, destRoot=thisRoot, recursive=opt.recursive, ignore=opt.exclude)
            except Exception as e:
                logger.error(str(e))
                logger.error("ignoring error")
        elif os.path.isfile(filename):
            if destIsDir:
                thisRoot = os.path.join(dest, os.path.basename(filename))
            copy(filename, thisRoot)
        else:
            logger.error("%s: No such file or directory." % (filename))


    logger.info("\nWaiting for transfers to complete.\nCTRL-\ to interrupt\n")

    queue.join()
    endTime = time.time()
    bytesSent = 0
    filesSent = 0
    bytesSkipped = 0
    filesSkipped = 0
    filesErrored = 0
    for stream in streams:
        bytesSent += stream.bytesSent
        bytesSkipped += stream.bytesSkipped
        filesSent += stream.filesSent
        filesSkipped += stream.filesSkipped
        filesErrored += stream.filesErrored

    logger.info("\n\n==== TRANSFER REPORT ====\n\n")

    if bytesSent > 0:
        rate = bytesSent / (endTime - startTime) / 1024.0
        logger.info("Sent %d files (%8.1f kbytes @ %8.3f kBytes/s)" % (filesSent, bytesSent / 1024.0, rate))
        speedUp = (bytesSkipped + bytesSent) / bytesSent
        logger.info("Speedup:  %f (skipped %d files)" % (speedUp, filesSkipped))

    if bytesSent == 0:
        logger.info("No files needed sending ")

    if filesErrored > 0:
        logger.info("Error transferring %d files, please try again" % (filesErrored))





"""Lists information about a VOSpace DataNode or the contents of a ContainerNode."""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import logging
import math
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception
import sys
import time
from .. import vos
from argparse import ArgumentError

# this is a pointer to the module object instance itself.
this = sys.modules[__name__]

# we can explicitly make assignments on it
this.human = False

__all__ = ['vls']


def size_format(size):
    """Format a size value for listing"""
    try:
        size = float(size)
    except Exception as ex:
        logging.debug(str(ex))
        size = 0.0
    if this.human:
        size_unit = ['B', 'K', 'M', 'G', 'T']
        # noinspection PyBroadException
        try:
            length = float(size)
            scale = int(math.log(length) / math.log(1024))
            length = "%.0f%s" % (length / (1024.0 ** scale), size_unit[scale])
        except:
            length = str(int(size))
    else:
        length = str(int(size))
    return "%12s " % length


def date_format(epoch):
    """given an epoch, return a unix-ls like formatted string"""

    time_tuple = time.localtime(epoch)
    if time.localtime().tm_year != time_tuple.tm_year:
        return time.strftime('%b %d  %Y ', time_tuple)
    return time.strftime('%b %d %H:%M ', time_tuple)


__LIST_FORMATS__ = {'permissions': lambda value: "{:<11}".format(value),
                    'creator': lambda value: " {:<20}".format(value),
                    'readGroup': lambda value: " {:<15}".format(value.replace(vos.CADC_GMS_PREFIX, "")),
                    'writeGroup': lambda value: " {:<15}".format(value.replace(vos.CADC_GMS_PREFIX, "")),
                    'isLocked': lambda value: " {:<8}".format("", "LOCKED")[value == "true"],
                    'size': size_format,
                    'date': date_format}

DESCRIPTION = """lists the contents of a VOSpace Node.

Long listing provides the file size, ownership and read/write status of Node.

"""


def vls():

    parser = CommonParser(description=DESCRIPTION, add_help=False)
    parser.add_argument('node', nargs="+", help="VOSpace Node to list.")
    parser.add_option("--help", action="help", default='==SUPPRESS==', help='show this help message and exit')
    parser.add_option("-l", "--long", action="store_true", help="verbose listing sorted by name")
    parser.add_option("-g", "--group", action="store_true", help="display group read/write information")
    parser.add_option("-h", "--human", action="store_true", help="make sizes human readable", default=False)
    parser.add_option("-S", "--Size", action="store_true", help="sort files by size", default=False)
    parser.add_option("-r", "--reverse", action="store_true", help="reverse the sort order", default=False)
    parser.add_option("-t", "--time", action="store_true", help="sort by time copied to VOSpace")

    try:
        opt = parser.parse_args()
        this.human = opt.human

        set_logging_level_from_args(opt)

        # set which columns will be printed
        columns = []
        if opt.long or opt.group:
            columns = ['permissions']
            if opt.long:
                columns.extend(['creator'])
            columns.extend(['readGroup', 'writeGroup', 'isLocked', 'size', 'date'])

        # determine if their is a sorting order
        sort_key = (opt.time and "date") or (opt.Size and "size") or False

        # create a client to send VOSpace command
        client = vos.Client(vospace_certfile=opt.certfile,
                            vospace_token=opt.token)

        for node in opt.node:
            if not node.startswith('vos:'):
                raise ArgumentError(opt.node, "Invalid node name: {}".format(node))
            logging.debug("getting listing of: %s" % str(node))
            info_list = client.get_info_list(node)

            if sort_key:
                # noinspection PyBroadException
                try:
                    sorted_list = sorted(info_list, key=lambda name: name[1][sort_key], reverse=not opt.reverse)
                except:
                    sorted_list = info_list
                finally:
                    info_list = sorted_list

            for item in info_list:
                name_string = item[0]
                for col in columns:
                    value = item[1].get(col, None)
                    value = value is not None and value or ""
                    if col in __LIST_FORMATS__:
                        sys.stdout.write(__LIST_FORMATS__[col](value))
                    if item[1]["permissions"][0] == 'l':
                        name_string = "%s -> %s" % (name_string, item[1]['target'])
                sys.stdout.write("%s\n" % name_string)
    except Exception as ex:
        exit_on_exception(ex)


vls.__doc__ = DESCRIPTION



#!python
"""remove a vospace data node, fails if container node or node is locked."""
import logging
from ..commonparser import set_logging_level_from_args, exit_on_exception, CommonParser
from .. import vos

DESCRIPTION = """remove a vospace data node; fails if container node or node is locked.

eg. vrm vos:/root/node   -- deletes a data node"""


def vrm():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument('node', help='dataNode or linkNode to delete from VOSpace', nargs='+')

    args = parser.parse_args()
    set_logging_level_from_args(args)

    try:
        client = vos.Client(vospace_certfile=args.certfile, vospace_token=args.token)
        for node in args.node:
            if not node.startswith('vos:'):
                logging.error("%s is not a valid VOSpace handle".format(node))
            if client.isfile(node) or client.get_node(node).islink():
                logging.info("deleting {}".format(node))
                client.delete(node)
            elif client.isdir(node):
                logging.error("{} is a directory".format(node))
            elif client.access(node):
                logging.info("deleting link {}".format(node))
                client.delete(node)
    except Exception as ex:
        exit_on_exception(ex)

vrm.__doc__ = DESCRIPTION



"""Delete a VOSpace ContainerNode (aka directory)"""
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception
import logging
from vos import vos

DESCRIPTION = """deletes a VOSpace container node (aka directory)

e.g.   vrmdir vos:Root/MyContainer

CAUTION:  The container need not be empty."""


def vrmdir():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument('nodes', help="Container nodes to delete from VOSpace", nargs='+')

    args = parser.parse_args()

    set_logging_level_from_args(args)

    try:
        client = vos.Client(vospace_certfile=args.certfile, vospace_token=args.token)
        for container_node in args.nodes:
            if not container_node.startswith("vos:"):
                raise ValueError("{} is not a valid VOSpace handle".format(container_node))
            if client.isdir(container_node):
                logging.info("deleting {}".format(container_node))
                client.delete(container_node)
            else:
                raise ValueError("{} is a not a container node".format(container_node))
    except Exception as ex:
        exit_on_exception(ex)


vrmdir.__doc__ = DESCRIPTION



"""Tags are annotations on VOSpace Nodes.  This module provides set/read/(list) functions for property(ies) of a node.

The tag system is meant to allow tags, in addition to the standard node properties. """
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
import pprint
from ..commonparser import CommonParser, set_logging_level_from_args, exit_on_exception
from .. import vos


DESCRIPTION = """provides set/read/(list) functions for property(ies) of a node.

Properties are attributes on the node.  There can be users attributes or system attributes.

Only user attributes can be set.

examples:

set at property:  vtag vos:RootNode/MyImage.fits quality=good
read a property:  vtag vos:RootNode/MyImage.fits quality
delete a property: vtag vos:RootNode/MyImage.fits quality=
                   or
                   vtag vos:RootNode/MyImage.fits quality --remove
list all property values:  vtag vos:RootNode/MyImage.fits

"""


def vtag():

    parser = CommonParser(description=DESCRIPTION)
    parser.add_argument('node', help='Node to set property (tag/attribute) on')
    parser.add_argument('property', help="Property whose value will be read, set or deleted", nargs="+")
    parser.add_option('--remove', action="store_true", help='remove the listed property')

    opt = parser.parse_args()
    args = opt
    set_logging_level_from_args(args)

    # the node should be the first argument, the rest should contain the key/val pairs
    node = args.node

    props = []
    if args.remove:
        # remove signified by blank value in key=value listing
        for prop in args.property:
            if '=' not in prop:
                prop += "="
            props.append(prop)
    else:
        props = args.property

    try:
        client = vos.Client(vospace_certfile=opt.certfile, vospace_token=opt.token)
        node = client.get_node(node)
        if len(props) == 0:
            # print all properties
            pprint.pprint(node.props)

        changed = False
        for prop in props:
            prop = prop.split('=')
            if len(prop) == 1:
                # get one property
                pprint.pprint(node.props.get(prop[0], None))
            elif len(prop) == 2:
                # update properties
                key, value = prop
                if len(value) == 0:
                    value = None
                if value != node.props.get(key, None):
                    node.props[key] = value
                    changed = True
            else:
                raise ValueError("Illegal keyword of value character ('=') used: %s" % ('='.join(prop)))

        if changed:
            client.add_props(node)
    except Exception as ex:
        exit_on_exception(ex)

vtag.__doc__ = DESCRIPTION



"""
Command line scripts for interacting with the VOSpace service.

These commands are designed to mimic their unix conterparts.
see:  command --help for details

"""
from .vcat import vcat
from .vchmod import vchmod
from .vcp import vcp
from .vln import vln
from .vlock import vlock
from .vls import vls
from .vmkdir import vmkdir
from .vmv import vmv
from .vrm import vrm
from .vrmdir import vrmdir
from .vsync import vsync
from .vtag import vtag

__all__ = ['vcp', 'vcat', 'vchmod', 'vcp', 'vln', 'vlock', 'vls', 'vmkdir', 'vmv', 'vrm', 'vrmdir', 'vsync', 'vtag']




""" keep track of vospace nodes that have been already been accessed during the current session."""
import threading
import logging

logger = logging.getLogger('vos')
#logger.setLevel(logging.ERROR)

class NodeCache(dict):
    """ A dictionary like object that provides the ability to look up a VOSpace nodes metadata.
     
    usage:
         # Create a node cache:
         nodeCache = NodeCache()

         with nodeCache.volatile(nodeURI):
            # Do things that might change the node value.
            # The cache will be cleared on entry

         with nodeCache.watch(nodeURI) as watch:
             # do things which shouldn't be cached when the node is
             # volatile.
             watch.insert(node)
             # The node will not be cached if the tree became volatile
             # at any point while the nodeURI was being watched.
    """

    def __init__(self, *args):
        """ Initialize the node cache."""
        dict.__init__(self, args)
        self.lock = threading.Lock()
        self.watched_nodes = []
        self.volatile_nodes = []

    def watch(self, uri):
        """Factory that returns a watch 2015 09 09.39169 for the given uri.
        
        :param uri: the VOSpace uri to watch
        """
        return self.Watch(self, uri.rstrip('/'))

    def volatile(self, uri):
        """Factory for volatile objects.
        
        :param uri: the VOSpace uri to tag as volatile
        """
        return self.Volatile(self, uri.rstrip('/'))

    def __missing__(self, key):
        """Attempting to access a non-cached node returns None rather than
           raising an exception.
        
        :param key: the key in the dict being accessed (i.e. uri)   
        """
        return None

    def __setitem__(self, key, value):
        """If an node is directly inserted into the cache, automatically create
           a watch.
        
        :param key: the uri of the node being inserted into the dictionary.
        :param value:  the actual node object that will be stored.
        """
        with self.watch(key) as w:
            w.insert(value)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.rstrip('/'))

    def __contains__(self, key):
        return dict.__contains__(self, key.rstrip('/'))

    class Volatile(object):
        """ Objects that mark a code segment where a uri is volatile and
            the Node in the cache shouldn't be used."""

        def __init__(self, node_cache, uri):
            """
            
            :param node_cache: the NOdeCache object this Volatile object is in. 
            :type node_cache: NodeCache
            :param uri: the VOSpace uri that references this node in the NodeCache
            :type uri: str
            """
            self.node_cache = node_cache
            self.uri = uri.rstrip('/')

        def __enter__(self):
            """ Mark any sub-trees being watched as being dirty
                add to self.nodeCache.volatileNodes.
                Remove any cached nodes in the volatile subtree.
            """

            with self.node_cache.lock:
                # Add this volatile object to a list of all active volatile objects.
                self.node_cache.volatile_nodes.append(self)

                # Remove any cached nodes in the volatile sub-tree.
                for uri in list(self.node_cache.keys()):
                    if uri.startswith(self.uri):
                        del self.node_cache[uri]

                # Clear the parent node as well to force an update
                parent = self.uri[:self.uri.rfind("/")]
                self.node_cache.pop(parent, None)

                # Mark any watched nodes in the volatile sub-tree dirty
                for watchedNode in self.node_cache.watched_nodes:
                    if watchedNode.uri.startswith(self.uri) or (watchedNode.uri == parent):
                        watchedNode.dirty = True

            return self

        def __exit__(self, exc_type, exc_value, traceback):
            """ Remove this volitile object from the list of active volatiles.
            """
            with self.node_cache.lock:
                self.node_cache.volatile_nodes.remove(self)

    class Watch(object):
        """ Objects that mark a code segment where a node has been read from
            vospace, and is intended to be cached.
        """

        def __init__(self, node_cache, uri):
            """
            
            :param node_cache: the NodeCache object containing the uri to watch.
             :type node_cache: NodeCache
            :param uri: the uri in the NodeCache that will be watched.
            :type uri: str
            
            """
            self.node_cache = node_cache
            self.uri = uri
            self.dirty = False

        def __enter__(self):
            with self.node_cache.lock:
                # Add this watch object to the list of active watch objects.
                self.node_cache.watched_nodes.append(self)

                # Check to see if this watch object is in an existing volatile
                # tree. If it is, mark this watch object as dirty.
                for this_volatile in self.node_cache.volatile_nodes:
                    if self.uri.startswith(this_volatile.uri):
                        self.dirty = True
                        return self
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            with self.node_cache.lock:
                self.node_cache.watched_nodes.remove(self)

        def insert(self, value):
            """ Insert an value, likely node object, into the cache, but only if the watch is not dirty."""
            if not self.dirty:
                # noinspection PyCallByClass
                dict.__setitem__(self.node_cache, self.uri, value)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import glob
import os
import sys
import imp
from setuptools.command.test import test as TestCommand
from setuptools import find_packages

from setuptools import setup

import distutils.cmd
import distutils.log
import subprocess

# read the README.rst file and return as string.
def readme():
    with open('README.rst') as r_obj:
        return r_obj.read()

# Get some values from the setup.cfg
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

conf = ConfigParser()
conf.optionxform=str
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('package_name', 'packagename')
DESCRIPTION = metadata.get('description', 'CADC package')
AUTHOR = metadata.get('author', 'CADC')
AUTHOR_EMAIL = metadata.get('author_email', 'cadc@nrc.gc.ca')
LICENSE = metadata.get('license', 'unknown')
URL = metadata.get('url', 'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca')

# VERSION should be PEP386 compatible (http://www.python.org/dev/peps/pep-0386)
VERSION = metadata.get('version', 'none')

# generate the version file
with open(os.path.join(PACKAGENAME, 'version.py'), 'w') as f:
    f.write('version = \'{}\''.format(VERSION))	

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']

# Define entry points for command-line scripts
entry_points = {'console_scripts': []}

entry_point_list = conf.items('entry_points')
for entry_point in entry_point_list:
    entry_points['console_scripts'].append('{0} = {1}'.format(entry_point[0],
                                                              entry_point[1]))

# add the --cov option to the test command
class PyTest(TestCommand):
    """class py.test for the testing

    """
    user_options = []

    def __init__(self, dist, **kw):
        TestCommand.__init__(self, dist, **kw)
        self.pytest_args = ['--cov', PACKAGENAME]

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        err_no = pytest.main(self.pytest_args)
        sys.exit(err_no)

# Note that requires and provides should not be included in the call to
# ``setup``, since these are now deprecated. See this link for more details:
# https://groups.google.com/forum/#!topic/astropy-dev/urYO8ckB2uM

setup(name=PACKAGENAME,
      version=VERSION,
      description=DESCRIPTION,
      scripts=scripts,
      install_requires=metadata.get('install_requires', '').strip().split(),
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      long_description=readme(),
      zip_safe=False,
      use_2to3=False,
      setup_requires=['pytest-runner'],
      entry_points=entry_points,
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, <4',
      packages=find_packages(),
      package_data={PACKAGENAME: ['data/*', 'tests/data/*', '*/data/*', '*/tests/data/*']},
      classifiers=[
        'Natural Language :: English',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
      ],
      cmdclass = {
          'coverage': PyTest,
      }
)



"""
A common commandline parser for the VOS command line tool set.
"""
import logging
import argparse
import os
import signal
import sys
from .version import version


# handle interrupts nicely
def signal_handler(signum, frame):
    raise KeyboardInterrupt("SIGINT signal handler. {0} {1}".format(signum, frame))

signal.signal(signal.SIGINT, signal_handler)


def exit_on_exception(ex):
    """
    Exit program due to an exception, print the exception and exit with error code.
    :param ex:
    :return:
    """
    logging.error(str(ex))
    sys.exit(getattr(ex, 'errno', -1)) if getattr(ex, 'errno', -1) else sys.exit(-1)


def set_logging_level_from_args(args):
    """Display version, set logging verbosity"""

    # Logger verbosity
    if args.debug:
        args.log_level = logging.DEBUG
    elif args.verbose:
        args.log_level = logging.INFO
    elif args.warning:
        args.log_level = logging.WARNING
    else:
        args.log_level = logging.ERROR

    log_format = "%(levelname)s %(module)s %(message)s"
    if args.log_level < logging.INFO:
        log_format = ("%(levelname)s %(asctime)s %(thread)d vos-" + str(version) +
                      " %(module)s.%(funcName)s.%(lineno)d %(message)s")
    logging.basicConfig(format=log_format, level=args.log_level)
    logger = logging.getLogger('root')
    logger.setLevel(args.log_level)

    if args.vos_debug:
        logger = logging.getLogger('vos')
        logger.setLevel(logging.DEBUG)

    if sys.version_info[1] > 6:
        logging.getLogger().addHandler(logging.NullHandler())


class CommonParser(argparse.ArgumentParser):
    """Class to hold and parse common command-line arguments for vos clients"""

    add_option = argparse.ArgumentParser.add_argument

    def __init__(self, *args, **kwargs):
        # call the parent constructor
        super(CommonParser, self).__init__(*args, formatter_class=argparse.RawDescriptionHelpFormatter,
                                           epilog="""Default service settings in ~/.config/vos/vos-config.
        """,
                                           **kwargs)
        # inherit the VOS client version
        self.version = version
        self.log_level = logging.ERROR

        # now add on the common parameters
        self.add_argument("--certfile",
                          help="filename of your CADC X509 authentication certificate",
                          default=os.path.join(os.getenv("HOME", "."),
                                               ".ssl/cadcproxy.pem"))
        self.add_argument("--token",
                          help="authentication token string (alternative to certfile)",
                          default=None)
        self.add_argument("--version", action="version",
                          version=version)
        self.add_argument("-d", "--debug", action="store_true", default=False,
                          help="print on command debug messages.")
        self.add_argument("--vos-debug", action="store_true", help="Print on vos debug messages.")
        self.add_argument("-v", "--verbose", action="store_true", default=False,
                          help="print verbose messages")
        self.add_argument("-w", "--warning", action="store_true", default=False,
                          help="print warning messages only")
